/*----------------------------------------------------------------------------
|   Copyright (c) 2000 Jochen Loewer (loewerj@hotmail.com)
|-----------------------------------------------------------------------------
|
|   $Header: $
|
| 
|   A (partial) XSLT implementation for tDOM, according to the W3C 
|   recommendation (16 Nov 1999).
|   See http://www.w3.org/TR/1999/REC-xslt-19991116 for details.
|
|
|   The contents of this file are subject to the Mozilla Public License
|   Version 1.1 (the "License"); you may not use this file except in
|   compliance with the License. You may obtain a copy of the License at
|   http://www.mozilla.org/MPL/
|
|   Software distributed under the License is distributed on an "AS IS"
|   basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
|   License for the specific language governing rights and limitations
|   under the License.
|
|   The Original Code is tDOM.
|
|   The Initial Developer of the Original Code is Jochen Loewer
|   Portions created by Jochen Loewer are Copyright (C) 1999, 2000
|   Jochen Loewer. All Rights Reserved.
|
|   Contributor(s):
|
|
|   $Log: $
|
|
|   written by Jochen Loewer
|   June, 2000
|
\---------------------------------------------------------------------------*/



/*----------------------------------------------------------------------------
|   Includes
|
\---------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include <dom.h>
#include <domxpath.h>
#include <domxslt.h>
#include <tcl.h>         /* for hash tables */


/*----------------------------------------------------------------------------
|   Defines
|
\---------------------------------------------------------------------------*/
#define XSLT_NAMESPACE  "http://www.w3.org/1999/XSL/Transform"
#define PRIO_UNKNOWN    -111.0


/*----------------------------------------------------------------------------
|   Macros
|
\---------------------------------------------------------------------------*/
#define DBG(x)            
#define TRACE(x)          DBG(fprintf(stderr,(x)))
#define TRACE1(x,a)       DBG(fprintf(stderr,(x),(a)))
#define TRACE2(x,a,b)     DBG(fprintf(stderr,(x),(a),(b)))
#define TRACE3(x,a,b,c)   DBG(fprintf(stderr,(x),(a),(b),(c)))
#define TRACE4(x,a,b,c,d) DBG(fprintf(stderr,(x),(a),(b),(c),(d)))

#define CHECK_RC          if (rc < 0) return rc
#define SET_TAG(t,n,s,v)  if (strcmp(n,s)==0) { t->info = v; return v; }


extern void printAst (int depth, ast t);

/*--------------------------------------------------------------------------
|   xsltTag
|
\-------------------------------------------------------------------------*/
typedef enum {

    unknown = 1,
    applyImports, applyTemplates, attribute, attributeSet, callTemplate, 
    choose, comment, copy, copyOf, decimalFormat, element, fallback, forEach,
    xsltIf, import,
    include, key, message, namespaceAlias, number, output, otherwise,
    param, procinstr,
    preserveSpace, sort, stylesheet, stripSpace, text, template,
    transform, valueOf, variable, when, withParam

} xsltTag;



/*--------------------------------------------------------------------------
|   xsltTemplate
|
\-------------------------------------------------------------------------*/
typedef struct xsltTemplate {

    char    * match;
    char    * name;
    ast       ast;
    char    * mode;
    double    prio;
    domNode * content;
        
    struct xsltTemplate *next;
    
} xsltTemplate;


/*--------------------------------------------------------------------------
|   xsltAttrSet
|
\-------------------------------------------------------------------------*/
typedef struct xsltAttrSet {

    char    * name;
    domNode * content;
        
    struct xsltAttrSet *next;
    
} xsltAttrSet;


/*--------------------------------------------------------------------------
|   xsltKeyValue
|
\-------------------------------------------------------------------------*/
typedef struct xsltKeyValue {

    char          * strValue;
    domNode       * node;

    struct xsltKeyValue * next;
    
} xsltKeyValue;

/*--------------------------------------------------------------------------
|   xsltKeyInfo
|
\-------------------------------------------------------------------------*/
typedef struct xsltKeyInfo {

    char          * name;
    char          * match;
    char          * use;
    int             evaluated;
    xsltKeyValue  * values;
    
    struct xsltKeyInfo * next;
    
} xsltKeyInfo;



/*--------------------------------------------------------------------------
|   xsltVariable
|
\-------------------------------------------------------------------------*/
typedef struct xsltVariable {

    char           * name;
    char           * select;
    domNode        * value;
    xpathResultSet   rs;
    
    struct xsltVariable *next;
        
} xsltVariable;


/*--------------------------------------------------------------------------
|   xsltVarFrame
|
\-------------------------------------------------------------------------*/
typedef struct xsltVarFrame {

    xsltVariable        * vars;
    struct xsltVarFrame * next;
        
} xsltVarFrame;


/*--------------------------------------------------------------------------
|   xsltState
|
\-------------------------------------------------------------------------*/
typedef struct {

    xsltTemplate      * templates;
    int                 nr_perserveSpace;
    char             ** perserveSpace;
    int                 nr_stripSpace;
    char             ** stripSpace;    
    domNode           * xmlRootNode;
    domDocument       * resultDoc;
    domNode           * lastNode;
    xsltVarFrame      * varFrames;
    xsltAttrSet       * attrSets;
    Tcl_HashTable       xpaths;
    xsltKeyInfo       * keyInfos;
    char              * outputMethod;
    char              * outputEncoding;
    char              * outputMediaType;
    xpathCBs            cbs;
    xpathFuncCallback   orig_funcCB;
    void              * orig_funcClientData;
    domNode           * current;
    
} xsltState;


/*--------------------------------------------------------------------------
|   Prototypes
|
\-------------------------------------------------------------------------*/
int ApplyTemplates ( xsltState *xs, xpathResultSet *nodeList, 
                     char *mode, char **errMsg);
                     
static int ExecActions (xsltState *xs, xpathResultSet *context,
                        domNode *currentNode, int currentPos, 
                        domNode *actionNode,  char **errMsg);



/*----------------------------------------------------------------------------
|   printXML
|
\---------------------------------------------------------------------------*/
static void printXML (domNode *node, int level, int maxlevel) {

    domTextNode *tnode;
    domProcessingInstructionNode *pi;
    char tmp[80];
    int  i, l;
    
    while (node) {

        for (i=0;i<level;i++) fprintf(stderr, "  ");
        if (node->nodeType == ELEMENT_NODE) {
            tmp[0] = '\0';
            if ((level>=maxlevel) && (node->firstChild)) {
                strcpy( tmp, "...");
            }
            fprintf(stderr, "<%s/%d> %s\n", node->nodeName, node->nodeNumber, tmp);
            if (level<maxlevel) {
                if (node->firstChild) printXML(node->firstChild, level+1, maxlevel);
            }
        }
        if (node->nodeType == TEXT_NODE) {
            tnode = (domTextNode*)node;
            l = tnode->valueLength;
            if (l > 70) l = 70;
            memmove(tmp, tnode->nodeValue, l);
            tmp[l] = '\0';
            fprintf(stderr, "'%s'\n", tmp);
        }
        if (node->nodeType == PROCESSING_INSTRUCTION_NODE) {
            pi = (domProcessingInstructionNode*)node;
            l = pi->targetLength;
            if (l > 70) l = 70;
            memmove(tmp, pi->targetValue, l);
            tmp[l] = '\0';
            fprintf(stderr, "<?%s ", tmp);
            l = pi->dataLength;
            if (l > 70) l = 70;
            memmove(tmp, pi->dataValue, l);
            tmp[l] = '\0';
            fprintf(stderr, "%s?>\n", tmp);
        }
        node = node->nextSibling;
    }
}


/*----------------------------------------------------------------------------
|   getAttr
|
\---------------------------------------------------------------------------*/
static char * getAttr (
    domNode  *node,
    char     *name
)
{
    domAttrNode *attr;
    char *localName, *nsName, prefix[MAX_PREFIX_LEN];
    
    attr = node->firstAttr; 
    while (attr) {
    
        nsName = domNamespaceURI((domNode*)attr);
        if (nsName && strcmp(nsName, XSLT_NAMESPACE)==0) {
            domSplitQName((char*)attr->nodeName, prefix, &localName);
            if (strcmp(localName,name)==0) {
                return attr->nodeValue;
            }
        } 
        attr = attr->nextSibling;
    }
    return NULL;
}


/*----------------------------------------------------------------------------
|   getTag
|
\---------------------------------------------------------------------------*/
static xsltTag getTag (
    domNode  *node
)
{
    char *name;
  
    if (node->nodeType != ELEMENT_NODE) {
        return unknown;
    }
    if (node->info != 0) {
        return (xsltTag)node->info;
    }
    name = domNamespaceURI(node);
    if ((name == NULL) || (strcmp(name, XSLT_NAMESPACE)!=0)) {
       node->info = (int)unknown;
       return unknown;
    }    
    name = domGetLocalName(node->nodeName);

    switch (*name) {
        case 'a': SET_TAG(node,name,"apply-imports",  applyImports);
                  SET_TAG(node,name,"apply-templates",applyTemplates);
                  SET_TAG(node,name,"attribute",      attribute);
                  SET_TAG(node,name,"attribute-set",  attributeSet);
                  break;
        case 'c': SET_TAG(node,name,"call-template",  callTemplate);
                  SET_TAG(node,name,"choose",         choose);
                  SET_TAG(node,name,"comment",        comment);
                  SET_TAG(node,name,"copy",           copy);
                  SET_TAG(node,name,"copy-of",        copyOf);
                  break;
        case 'd': SET_TAG(node,name,"decimal-format", decimalFormat);
                  break;
        case 'e': SET_TAG(node,name,"element",        element);
                  break;
        case 'f': SET_TAG(node,name,"fallback",       fallback);
                  SET_TAG(node,name,"for-each",       forEach);
                  break;
        case 'i': SET_TAG(node,name,"if",             xsltIf);
                  SET_TAG(node,name,"import",         import);
                  SET_TAG(node,name,"include",        include);
                  break;
        case 'k': SET_TAG(node,name,"key",            key);
                  break;
        case 'm': SET_TAG(node,name,"message",        message);
                  break;
        case 'n': SET_TAG(node,name,"namespace-alias",namespaceAlias);
                  SET_TAG(node,name,"number",         number);
                  break;
        case 'o': SET_TAG(node,name,"output",         output);
                  SET_TAG(node,name,"otherwise",      otherwise);
                  break;
        case 'p': SET_TAG(node,name,"param",          param);
                  SET_TAG(node,name,"preserve-space", preserveSpace);
                  SET_TAG(node,name,"processing-instruction", procinstr);
                  break;
        case 's': SET_TAG(node,name,"sort",           sort);
                  SET_TAG(node,name,"stylesheet",     stylesheet);
                  SET_TAG(node,name,"strip-space",    stripSpace);
                  break;
        case 't': SET_TAG(node,name,"template",       template);
                  SET_TAG(node,name,"text",           text);
                  SET_TAG(node,name,"transform",      transform);
                  break;
        case 'v': SET_TAG(node,name,"value-of",       valueOf);
                  SET_TAG(node,name,"variable",       variable);
                  break;
        case 'w': SET_TAG(node,name,"when",           when);
                  SET_TAG(node,name,"with-param",     withParam);
                  break;
    } 
    node->info = (int)unknown;
    return unknown;    
}


/*----------------------------------------------------------------------------
|   xsltPopVarFrame
|
\---------------------------------------------------------------------------*/
static void xsltPopVarFrame (
    xsltState  * xs
)
{
    xsltVarFrame  * currentFrame;
    xsltVariable  * var, *vtmp;
    
    if (xs->varFrames) {    
        currentFrame = xs->varFrames;        
        xs->varFrames = currentFrame->next;
        /* free vars */
        var = currentFrame->vars;
        while (var) {
            free(var->name);
            vtmp = var;
            var = var->next;
            free(vtmp);
        }
        free(currentFrame);           
    }
}


/*----------------------------------------------------------------------------
|   xsltPushVarFrame
|
\---------------------------------------------------------------------------*/
static void xsltPushVarFrame (
    xsltState  * xs
)
{
    xsltVarFrame  * currentFrame;
    
    currentFrame = (xsltVarFrame*) malloc(sizeof(xsltVarFrame));
    currentFrame->vars = NULL;
    currentFrame->next = xs->varFrames;

    xs->varFrames = currentFrame;
}


/*----------------------------------------------------------------------------
|   xsltXPathFuncs
|
\---------------------------------------------------------------------------*/
static int xsltXPathFuncs (
    void            * clientData,
    char            * funcName,
    domNode         * ctxNode,
    int               ctxPos,
    xpathResultSet  * ctx,
    int               argc,
    xpathResultSets * argv,
    xpathResultSet  * result,
    char           ** errMsg
)
{
    xsltState     * xs = clientData;
    xsltKeyInfo   * kinfo;
    domNode       * node;
    char          * keyId, *filterValue, *useValue;
    ast             matchAST, useAST;
    int             first, rc;
    xpathResultSet  rs, context;
    xsltKeyValue  * value = NULL;
    
    /* fprintf(stderr,"xsltXPathFuncs funcName='%s'\n",funcName); */
 
    if (strcmp(funcName, "key")==0) {
        DBG(fprintf(stderr,"xslt key function called!\n");)
        if (argc != 2) {
            *errMsg = "xsl:key needs two arguments!";
            return 1;
        }
        keyId = xpathFuncString(argv[0]);
        TRACE1("keyId='%s' \n", keyId);
        kinfo = xs->keyInfos;
        while (kinfo) {
            if (strcmp(keyId, kinfo->name)==0) {
                TRACE1("keyInfo for keyId (%s) found!\n", keyId);
                break;
            }
            kinfo = kinfo->next;
        }
        if (!kinfo) {
            *errMsg = "key information not found!\n";
            return 1;
        }
        if (!kinfo->evaluated) {
            /*---------------------------------------------------------
            |   build up key/index information once!
            |   i.e. scan through complete source DOM tree,
            |   check for the matching criteria and if a match 
            |   was found extrace 'use' information
            | 
            \--------------------------------------------------------*/
            rc = xpathParse(kinfo->match, errMsg, &matchAST, 1);
            CHECK_RC;
            TRACE1("parsing use '%s' \n", kinfo->use);
            rc = xpathParse(kinfo->use, errMsg, &useAST, 0);
            CHECK_RC;
            node = xs->xmlRootNode;
            first = 1;
            while (node && (first || node != xs->xmlRootNode)) {
                rc = xpathMatches ( matchAST, node, &(xs->cbs), errMsg);
                if (rc < 0) {
                    TRACE1("xpathMatches had errors '%s' \n", *errMsg);
                    return rc;
                }
                if (rc > 0) {
                    TRACE("found match for key !\n");
                    xpathRSInit( &rs );
                    xpathRSInit( &context );
                    rsAddNode( &context, node );
                    printXML(node, 0, 2);
                    rc = xpathEvalSteps( useAST, &context, node, 0,
                                         &(xs->cbs), &rs, errMsg);
                    if (rc != XPATH_OK) {
                        xpathRSFree( &rs );
                        return rc;
                    }
                    useValue = xpathFuncString( &rs );
                    TRACE1("use value = '%s'\n", useValue);
                    value = malloc(sizeof(xsltKeyValue));
                    value->strValue = useValue;
                    value->node     = node;
                    value->next     = kinfo->values;
                    kinfo->values = value;
                    xpathRSFree( &context );
                    xpathRSFree( &rs );                   
                }
                if ((node->nodeType == ELEMENT_NODE) && (node->firstChild)) {
                    node = node->firstChild;
                    first = 0;
                    continue;
                }
                if (node->nextSibling) {
                    node = node->nextSibling;
                    continue;
                }
                while ( node->parentNode &&
                       (node->parentNode != xs->xmlRootNode) &&
                       (node->parentNode->nextSibling == NULL) ) {

                    node = node->parentNode;
                }
                if ((node != xs->xmlRootNode) &&
                    (node->parentNode)        &&
                    (node->parentNode != xs->xmlRootNode)
                ) {
                    node = node->parentNode->nextSibling;
                } else {
                    break;
                }
            }
        }
        if (argv[1]->type == NodeSetResult) {
        } else {
           filterValue = xpathFuncString(argv[1]);
           TRACE1("filterValue='%s' \n", filterValue); 
           value = kinfo->values;
           while (value) {
               TRACE2("checking '%s' with '%s' \n", filterValue, value->strValue);
               if (strcmp(filterValue, value->strValue)==0) {
                   break;
               }
               value = value->next;
           }
           if (value) {
               rsAddNode(result, value->node);
           }
           return 0;
        } 
    } else
    if (strcmp(funcName, "current")==0) {
        DBG(fprintf(stderr, "xsltXPathFuncs 'current' = '%d' \n", xs->current->nodeNumber);)
        rsAddNode(result, xs->current);
        return 0;
    } else {
        /* chain back to original callback */
        if (xs->orig_funcCB) {
            return (xs->orig_funcCB)(xs->orig_funcClientData, funcName,
                                     ctxNode, ctxPos, ctx, argc, argv,
                                     result, errMsg);
        }
    }
    return 0;    
}



/*----------------------------------------------------------------------------
|   evalXPath
|
\---------------------------------------------------------------------------*/
static int evalXPath (
    xsltState       * xs,
    xpathResultSet  * context,
    domNode         * currentNode,
    int               currentPos,
    char            * xpath,    
    xpathResultSet  * rs,
    char           ** errMsg
)
{
    int rc, hnew;
    ast t;
    Tcl_HashEntry *h;

    h = Tcl_CreateHashEntry (&(xs->xpaths), xpath, &hnew);
    if (!hnew) {
        t = (ast)Tcl_GetHashValue(h);
    } else {     
        rc = xpathParse(xpath, errMsg, &t, 0);
        CHECK_RC;
        Tcl_SetHashValue(h, t);
    }
    xpathRSInit( rs );
    
    DBG(printAst(0,t);)
    rc = xpathEvalSteps( t, context, currentNode, currentPos,
                         &(xs->cbs), rs, errMsg); 
    if (rc != XPATH_OK) {
        xpathRSFree( rs );
    } 
                            
    /* xpathFreeAst( t ); 
    */
    return rc;
}



/*----------------------------------------------------------------------------
|   nodeGreater
|
\---------------------------------------------------------------------------*/
static int nodeGreater (
    xsltState * xs,
    ast         t,
    char      * select,
    int         typeText,
    int         asc,
    domNode   * a,
    domNode   * b, 
    char     ** errMsg,
    int       * greater
)
{
    xpathResultSet  inA, inB, rsA, rsB;
    int             rc, NaN;
    char           *strA, *strB;
    double          realA, realB;

    *greater = 0;

    xpathRSInit(&inA   ); xpathRSInit( &rsA );
    rsAddNode  (&inA, a);  

    xpathRSInit(&inB   ); xpathRSInit( &rsB );
    rsAddNode  (&inB, b);  
 
    rc = xpathEvalSteps( t, &inA, a, 0, &(xs->cbs), &rsA, errMsg); 
    if (rc != XPATH_OK) {
        xpathRSFree( &inA ); xpathRSFree( &rsA );
        xpathRSFree( &inB ); xpathRSFree( &rsB );
        return rc;
    }                         
    rc = xpathEvalSteps( t, &inB, b, 0, &(xs->cbs), &rsB, errMsg); 
    if (rc != XPATH_OK) {
        xpathRSFree( &inA ); xpathRSFree( &rsA );
        xpathRSFree( &inB ); xpathRSFree( &rsB );
        return rc;
    }

    if (typeText) {
        strA = xpathFuncString( &rsA );
        strB = xpathFuncString( &rsB );
        
        rc = strcmp(strA, strB);
        if (asc) *greater = (rc > 0);
            else *greater = (rc < 0);

        free(strA); free(strB);       
    } else {
        realA = xpathFuncNumber( &rsA, &NaN );
        realB = xpathFuncNumber( &rsB, &NaN );
        if (asc) *greater = (realA > realB); 
            else *greater = (realA < realB);        
    }
    
    xpathRSFree( &inA ); xpathRSFree( &rsA );
    xpathRSFree( &inB ); xpathRSFree( &rsB );
    return 0;
}


/*----------------------------------------------------------------------------
|   sortNodeSetInsert  -  use insertion sort for keeping sequence
|                         in pre/sub-sorted lists
|
\---------------------------------------------------------------------------*/
static int sortNodeSetInsert(
    xsltState * xs,
    ast         t,
    char      * sel,
    int         txt,
    int         asc,
    domNode   * nodes[], 
    int         n,
    char     ** errMsg
)
{
    domNode *tmp;
    int rc, gt, a, b;

    for (a = 1; a < n; a++) {
        tmp = nodes[a];
        b = a - 1;
        while (b >= 0) {
            rc = nodeGreater(xs,t,sel,txt,asc,nodes[b],tmp,errMsg,&gt);
            CHECK_RC;
            if (!gt) break;
            nodes[b+1] = nodes[b];
            b--;
        }
        nodes[b+1] = tmp;
    }
    return 0;
}


/*----------------------------------------------------------------------------
|   sortNodeSetQuick  -   use quicksort of fast sorting
|
\---------------------------------------------------------------------------*/
static int sortNodeSetQuick(
    xsltState * xs,
    ast         t,
    char      * sel,
    int         txt,
    int         asc,
    domNode   * nodes[], 
    int         n,
    char     ** errMsg
)
{
    domNode *tmp;
    int rc, gt, i, j, ln, rn;

    while (n > 1) {
        tmp = nodes[0]; nodes[0] = nodes[n/2]; nodes[n/2] = tmp;
        for (i = 0, j = n; ; ) {
            gt = 1;
            while (gt) {
                j--;
                rc = nodeGreater(xs,t,sel,txt,asc,nodes[j],nodes[0],errMsg,&gt);
                CHECK_RC;
            }
            gt = 1;
            while (i < j && gt) {
                i++; 
                rc = nodeGreater(xs,t,sel,txt,asc,nodes[0],nodes[i],errMsg,&gt);
                CHECK_RC;
            }
            if (i >= j)  break;
            tmp = nodes[i]; nodes[i] = nodes[j]; nodes[j] = tmp;
        }
        tmp = nodes[j]; nodes[j] = nodes[0]; nodes[0] = tmp;
        ln = j;
        rn = n - ++j;
        if (ln < rn) {
            rc = sortNodeSetQuick(xs, t, sel, txt, asc, nodes, ln, errMsg);
            CHECK_RC;
            nodes += j;
            n = rn;
        } else {
            rc = sortNodeSetQuick(xs, t, sel, txt, asc, &(nodes[j]), rn, errMsg);
            CHECK_RC;
            n = ln;
        }
    }
    return 0;
}


/*----------------------------------------------------------------------------
|   sortNodeSetByNodeNumber
|
\---------------------------------------------------------------------------*/
static void sortNodeSetByNodeNumber(
    domNode *nodes[], 
    int      n
)
{
    domNode *tmp;
    int i, j, ln, rn;

    while (n > 1) {
        tmp = nodes[0]; nodes[0] = nodes[n/2]; nodes[n/2] = tmp;
        for (i = 0, j = n; ; ) {
            do --j; while (nodes[j]->nodeNumber > nodes[0]->nodeNumber);
            do ++i; while (i < j && nodes[0]->nodeNumber > nodes[i]->nodeNumber);
            if (i >= j)  break;
            tmp = nodes[i]; nodes[i] = nodes[j]; nodes[j] = tmp;
        }
        tmp = nodes[j]; nodes[j] = nodes[0]; nodes[0] = tmp;
        ln = j;
        rn = n - ++j;
        if (ln < rn) {
            sortNodeSetByNodeNumber(nodes, ln);
            nodes += j;
            n = rn;
        } else {
            sortNodeSetByNodeNumber(&(nodes[j]), rn);
            n = ln;
        }
    }
}


/*----------------------------------------------------------------------------
|   sortByDocOrder
|
\---------------------------------------------------------------------------*/
static void sortByDocOrder (
    xpathResultSet  * rs
)
{
    if (rs->type != NodeSetResult) return;
    sortNodeSetByNodeNumber(rs->nodes, rs->nr_nodes);
}


/*----------------------------------------------------------------------------
|   xsltSetVar
|
\---------------------------------------------------------------------------*/
static int xsltSetVar (
    xsltState       * xs,
    int               forNextLevel,
    char            * variableName,
    xpathResultSet  * context,
    domNode         * node,
    int               currentPos,
    char            * select,
    domNode         * actionNodeChild,
    char           ** errMsg
)
{
    xsltVariable   * var;
    int              rc;
    xpathResultSet   rs;
    xsltVarFrame    *tmpFrame = NULL;
    domNode         *fragmentNode, *savedLastNode;
        

    TRACE1("xsltSetVar variableName='%s' \n", variableName);
    if (select!=NULL) {
        if (forNextLevel) {
            /* hide new variable frame */
            tmpFrame = xs->varFrames;
            xs->varFrames = tmpFrame->next;
        }
        TRACE2("xsltSetVar variableName='%s' select='%s'\n", variableName, select);
        xs->current = node;    
        rc = evalXPath (xs, context, node, currentPos, select, &rs, errMsg);
        CHECK_RC;
        if (forNextLevel) {
            /* show new variable frame again */
            xs->varFrames = tmpFrame;
        }
    } else {
        fragmentNode = domNewElementNode(xs->resultDoc, "(fragment)", ELEMENT_NODE);
        savedLastNode = xs->lastNode;
        xs->lastNode = fragmentNode;
        /* process the children as well */
        rc = ExecActions(xs, context, node, currentPos, actionNodeChild, errMsg);
        CHECK_RC;
        xpathRSInit(&rs);
        rsAddNode(&rs, fragmentNode);
        xs->lastNode = savedLastNode;
       /*                                                  
        xpathRSInit(&rs);
        while (node) {
            rsAddNode(&rs, node);
            node = node->nextSibling;
        }
       */
    }
    var = NULL;   
    if (!xs->varFrames) {    
        var = xs->varFrames->vars;
        while (var) {
            if (strcmp(var->name, variableName)==0) {
                break; /* found the variable */
            }
            var = var->next;
        }
    }
    if (!var) {
        var = (xsltVariable*) malloc(sizeof(xsltVariable));
        var->name = strdup(variableName);
        var->next = xs->varFrames->vars;
        xs->varFrames->vars = var;        
    }        
    var->select = select;
    var->value  = node;
    var->rs     = rs;
    DBG(rsPrint(&(var->rs)));
    return 0;
}


/*----------------------------------------------------------------------------
|   xsltVarExists
|
\---------------------------------------------------------------------------*/
static int xsltVarExists (
    xsltState  * xs,
    char       * variableName
)
{
    xsltVariable   * var;
        

    if (!xs->varFrames) return 0;
    
    var = xs->varFrames->vars;
    while (var) {
        if (strcmp(var->name, variableName)==0) {
            break; /* found the variable */
        }
        var = var->next;
    }
    if (var) return 1;
    return 0;
}


/*----------------------------------------------------------------------------
|   xsltGetVar
|
\---------------------------------------------------------------------------*/
static int xsltGetVar (
    void           * clientData,
    char           * variableName,
    xpathResultSet * result,
    char           **errMsg
)
{
    xsltState    * xs = clientData;
    xsltVarFrame * frame;
    xsltVariable * var;
    int            d; 
       
#if 0       
    TRACE1("xsltGetVar variableName='%s' \n", variableName);
#endif    

    frame = xs->varFrames;
    d = 0;
    while (frame) {
        var = frame->vars;
        while (var) {
            /*TRACE2("is it var %d:'%s' ? \n", d, var->name);*/
            if (strcmp(var->name, variableName)==0) {
                TRACE1("xsltGetVar '%s':\n", variableName);
                DBG(rsPrint(&(var->rs)));
                rsCopy(result, &(var->rs) );
#if 0                
                if (var->select) {
                } else {
                    TRACE1("rsAddNode  variableName='%s' \n", variableName);
                    rsAddNode(result, var->value);  
                }
#endif                
                return XPATH_OK;
            }
            var = var->next;
        }
        frame = frame->next;
        d++;
    }
    rsSetString(result, "");
    return XPATH_OK;
}



/*----------------------------------------------------------------------------
|   xsltAddTemplate
|
\---------------------------------------------------------------------------*/
static int xsltAddTemplate (
    xsltState *xs,    
    domNode   *node,
    char     **errMsg
)
{
    xsltTemplate *tpl;
    char         *prioStr;
    int           rc;
    
    tpl = malloc(sizeof(xsltTemplate));
    
    tpl->match   = getAttr(node,"match");
    tpl->name    = getAttr(node,"name");  
    tpl->ast     = NULL;
    tpl->mode    = getAttr(node,"mode"); 
    tpl->prio    = 0.5;
    tpl->content = node;
    tpl->next    = xs->templates;

    prioStr      = getAttr(node,"prio");
    if (prioStr) {
        tpl->prio = (double)atof(prioStr);
    }
    
    TRACE1("compiling XPATH '%s' ...\n", tpl->match);
    if (tpl->match) {
        
        rc = xpathParse(tpl->match, errMsg, &(tpl->ast), 1); 
        CHECK_RC;
        if (!prioStr) {
            tpl->prio = xpathGetPrio(tpl->ast);
        }
        DBG(printAst( 0, tpl->ast);)
    }
    TRACE4("AddTemplate '%s' '%s' '%s' '%2.2f' \n\n",
            tpl->match, tpl->name, tpl->mode, tpl->prio);

    xs->templates = tpl;
    return 0;           
}


/*----------------------------------------------------------------------------
|   ExecUseAttributeSets
|
\---------------------------------------------------------------------------*/
int ExecUseAttributeSets (
    xsltState         * xs,
    xpathResultSet    * context,
    domNode           * currentNode,
    int                 currentPos,
    char              * styles,
    char             ** errMsg
)
{
    xsltAttrSet *attrSet;
    char        *pc, *aSet, save;
    int          rc;
        
    pc = styles;
    while (*pc == ' ') pc++;
    aSet = pc;
                    
    while (*pc) {
        while (*pc && (*pc != ' ')) pc++;
        save = *pc;
        *pc = '\0';
        TRACE1("use-attribute-set '%s' \n", aSet);
        attrSet = xs->attrSets;
        while (attrSet) {
            TRACE2("use-Attr: '%s' == '%s' ? \n", attrSet->name, aSet);
            if (strcmp(attrSet->name, aSet)==0) {
                rc = ExecActions(xs, context, currentNode, currentPos, 
                                 attrSet->content->firstChild, errMsg);
                CHECK_RC;
                break;              
            }  
            attrSet = attrSet->next;
        }
        *pc = save;
        while (*pc == ' ') pc++;
        aSet = pc;
    }
    return 0;           
}



/*----------------------------------------------------------------------------
|   evalAttrTemplates
|
\---------------------------------------------------------------------------*/
static int evalAttrTemplates (
    xsltState       * xs,
    xpathResultSet  * context,
    domNode         * currentNode,
    int               currentPos,
    char            * str,
    char           ** out,
    char           ** errMsg
)
{
    xpathResultSet  rs;
    char           *tplStart = NULL, *tplResult, *pc;
    int             rc, aLen, inTpl = 0, p = 0;
    
    aLen = 500;
    *out = malloc(aLen);
    while (*str) {
        if (inTpl) {
            if (*str == '}') {
            
                *str = '\0';
                TRACE1("attrTpl: '%s' \n", tplStart);
                xs->current = currentNode;    
                rc = evalXPath (xs, context, currentNode, currentPos,
                                tplStart, &rs, errMsg);
                *str = '}';
                CHECK_RC;
                tplResult = xpathFuncString( &rs );
                DBG(fprintf(stderr, "attrTpl tplResult='%s' \n", tplResult);)
                xpathRSFree( &rs );
                pc = tplResult;
                while (*pc) {
                   (*out)[p++] = *pc++;
                    if (p>=aLen) { /* enlarge output buffer */
                         *out = realloc(*out, 2*aLen);
                         aLen += aLen;
                    }
                }
                inTpl = 0;
                free(tplResult);
            }
        } else {
            if (*str == '{') {
                if (*(str+1) == '{') {
                    /*-----------------------------------------------------
                    |    read over escaped '{':
                    |        '{{text text}}' -> '{text text}'
                    \----------------------------------------------------*/
                    str++;
                    (*out)[p++] = *str++;
                    if (p>=aLen) { /* enlarge output buffer */
                        *out = realloc(*out, 2*aLen);
                        aLen += aLen;
                    }
                    while (*str && (*str != '}') && (*(str-1) != '}')) {
                        (*out)[p++] = *str++;
                        if (p>=aLen) { /* enlarge output buffer */
                            *out = realloc(*out, 2*aLen);
                            aLen += aLen;
                        }
                    }
                } else {
                    tplStart = str+1;
                    inTpl = 1;
                }
            } else {
                (*out)[p++] = *str;
                if (p>=aLen) { /* enlarge output buffer */
                    *out = realloc(*out, 2*aLen);
                    aLen += aLen;
                }
            }
        }
        str++;
    }
    (*out)[p] = '\0';
    DBG(fprintf(stderr, "evalAttrTemplates out='%s' \n", (*out) );)
    return 0;
}


/*----------------------------------------------------------------------------
|   setParamVars
|
\---------------------------------------------------------------------------*/
static int setParamVars (
    xsltState       * xs,
    xpathResultSet  * context,
    domNode         * currentNode,
    int               currentPos,
    domNode         * actionNode,
    char           ** errMsg
)
{
    domNode *child;
    char    *str, *select;
    int      rc;
    
    child = actionNode->firstChild;
    while (child) {
        if (child->nodeType == ELEMENT_NODE) {
            TRACE1("setParamVars child '%s' \n", child->nodeName);
            if (getTag(child) == withParam) {
                str = getAttr(child, "name");
                if (str) {
                    TRACE1("setting with-param '%s' \n", str);
                    select = getAttr(child, "select");
                    if (select) {
                        TRACE1("with-param select='%s'\n", select);
                        rc = xsltSetVar(xs, 1, str, context, currentNode, currentPos, 
                                        select, NULL, errMsg);
                    } else {
                        rc = xsltSetVar(xs, 1, str, context, currentNode, currentPos, 
                                        NULL, child->firstChild, errMsg);
                    }
                    CHECK_RC;
                }
            }
        }
        child = child->nextSibling;
    }
    return 0;
}


/*----------------------------------------------------------------------------
|   doSortActions
|
\---------------------------------------------------------------------------*/
static int doSortActions (
    xsltState       * xs,
    xpathResultSet  * nodelist,
    domNode         * actionNode,
    int             * sorted,
    char           ** errMsg    
)
{
    domNode *child;
    char    *str, *select, *lang;
    int      rc, typeText, ascending, upperFirst = 1;
    ast      t;
    
    *sorted = 0;
    if (nodelist->type != NodeSetResult) return 0;
    
    child = actionNode->lastChild; /* do it backwards, so that multiple sort
                                      levels are correctly processed */
    while (child) {
        if (child->nodeType == ELEMENT_NODE) {
            TRACE1("doSortActions child '%s' \n", child->nodeName);
            if (getTag(child) == sort) {
                typeText  = 1;
                ascending = 1;
                select = getAttr(child, "select");
                if (!select) select = ".";
                str = getAttr(child, "data-type");
                if (str) {
                    if (strcmp(str,"number")==0) typeText = 0;
                }
                str = getAttr(child, "order");
                if (str) {
                    if (strcmp(str,"descending")==0) ascending = 0;
                }
                /* jcl: TODO */
                str = getAttr(child, "case-order");
                if (str) {
                    if (strcmp(str,"lower-first")==0) upperFirst = 0;
                }
                /* jcl: TODO */
                lang = getAttr(child, "lang");
                
                TRACE3("sorting with '%s' typeText %d ascending %d\n", 
                       select, typeText, ascending);
                rc = xpathParse(select, errMsg, &t, 0);
                CHECK_RC;
                if (*sorted) {
                    rc = sortNodeSetInsert(xs, t, select, typeText, ascending,
                            nodelist->nodes, nodelist->nr_nodes, errMsg);                                     
                } else {
                    rc = sortNodeSetQuick(xs, t, select, typeText, ascending,
                             nodelist->nodes, nodelist->nr_nodes, errMsg);
                }
                CHECK_RC;
                xpathFreeAst( t );                 
                *sorted = 1;
            }
        }
        child = child->previousSibling;
    }
    return 0;
}


/*----------------------------------------------------------------------------
|   xsltNumber
|
\---------------------------------------------------------------------------*/
static int xsltNumber (
    xsltState       * xs,
    char            * value,
    char            * level,
    char            * count,
    char            * format,
    xpathResultSet  * context,
    domNode         * currentNode,
    int               currentPos,
    char           ** errMsg
)
{
    xpathResultSet  rs;
    int             rc, v, NaN;
    char            tmp[80];
    
    TRACE3("xsltNumber value='%s' format='%s' count='%s' \n", value,format,count);

    v = currentPos+1;
    if (value) {
        xs->current = currentNode;
        rc = evalXPath(xs, context, currentNode, currentPos,
                       value, &rs, errMsg);
        CHECK_RC;
        v = xpathFuncNumber( &rs, &NaN );
        xpathRSFree( &rs );     
    }
    if (strcmp(format,"1")==0) {
        sprintf(tmp, "%d", v);
        domAppendNewTextNode(xs->lastNode, tmp, strlen(tmp), TEXT_NODE);
    } else {
    }
    return 0;
}



/*----------------------------------------------------------------------------
|   ExecAction
|
\---------------------------------------------------------------------------*/
static int ExecAction (
    xsltState       * xs,
    xpathResultSet  * context,
    domNode         * currentNode,
    int               currentPos,
    domNode         * actionNode,
    char           ** errMsg
)
{
    domNode        *child, *n, *savedLastNode;
    xsltTemplate   *tpl;
    domAttrNode    *attr;
    domTextNode    *tnode;
    domProcessingInstructionNode *pi;      
    xpathResultSet  rs, nodeList;
    char           *str, *str2, *mode, *select, *pc, tmp[80];
    char           *nsAT, *ns, *value, *level, *count, *format, *disableEsc;
    char           *uri, *localName, prefix[MAX_PREFIX_LEN];
    int             rc, b, i, len, sorted;
    

    if (actionNode->nodeType == TEXT_NODE) {
        domAppendNewTextNode(xs->lastNode, 
                             ((domTextNode*)actionNode)->nodeValue,
                             ((domTextNode*)actionNode)->valueLength, TEXT_NODE);
        return 0;
    }
    if (actionNode->nodeType != ELEMENT_NODE) return 0;

    TRACE1("\nExecAction '%s' \n", actionNode->nodeName);
    switch ( getTag(actionNode) ) {
    
        case applyImports: return 0;

        case applyTemplates:
            mode   = getAttr(actionNode, "mode");
            select = getAttr(actionNode, "select");
            if (!select) { 
                strcpy(tmp, "*|text()");
                select = tmp;
            }
            xpathRSInit( &nodeList );
            rsAddNode( &nodeList, currentNode );
            DBG(rsPrint( &nodeList ));
            TRACE2("applyTemplates: select = '%s' mode='%s'\n", select, mode);
            xs->current = currentNode;
            rc = evalXPath(xs, &nodeList, NULL, 0, select, &rs, errMsg);
            CHECK_RC;
            TRACE1("applyTemplates: evalXPath for select = '%s' gave back:\n", select);
            DBG(rsPrint(&rs));
            
            rc = doSortActions (xs, &rs, actionNode, &sorted, errMsg);
            CHECK_RC;
            if (!sorted) sortByDocOrder(&rs);

            TRACE1("                evalXPath for select = '%s': (SORTED)\n", select);
            DBG(rsPrint(&rs));

            rc = ApplyTemplates(xs, &rs, mode, errMsg); 
            CHECK_RC;
            xpathRSFree( &rs ); 
            xpathRSFree( &nodeList ); 
            break;
            
        case attribute:
            nsAT = getAttr(actionNode, "namespace");
            str = getAttr(actionNode, "name");
            if (str) {                
                rc = evalAttrTemplates( xs, context, currentNode, currentPos,
                                        str, &str2, errMsg);
                CHECK_RC;
                ns = NULL;
                if (nsAT) {
                    rc = evalAttrTemplates( xs, context, currentNode, currentPos,
                                            nsAT, &ns, errMsg);
                    CHECK_RC;
                }
                savedLastNode = xs->lastNode;
                xs->lastNode  = domNewElementNode (xs->resultDoc,  
                                    "container", ELEMENT_NODE);
                rc = ExecActions(xs, context, currentNode, currentPos,
                                 actionNode->firstChild, errMsg);
                CHECK_RC;
                pc = xpathGetTextValue (xs->lastNode, &len);
                domSetAttributeNS (savedLastNode, str2, pc, ns);
                free(pc);
                domFreeNode (xs->lastNode, NULL, NULL);
                xs->lastNode = savedLastNode;
                free(str2);
            }
            break;
            
        case attributeSet: return 0;
            
        case callTemplate:
            str = getAttr(actionNode, "name");
            if (str) {
               for( tpl = xs->templates; tpl != NULL; tpl = tpl->next) {
                   if (tpl->name && (strcmp(tpl->name,str)==0)) {
                       xsltPushVarFrame (xs);
                       TRACE2("call template %s match='%s' \n", str, tpl->match);
                       DBG(printXML(xs->lastNode, 0, 2);)
                       rc = setParamVars (xs, context, currentNode, currentPos, actionNode, errMsg);
                       CHECK_RC; 
                       rc = ExecActions(xs, context, currentNode, currentPos, 
                                        tpl->content->firstChild, errMsg);
                       TRACE2("called template '%s': ApplyTemplate/ExecActions rc = %d \n", str, rc);
                       CHECK_RC;    
                       xsltPopVarFrame (xs);
                       DBG(printXML(xs->lastNode, 0, 2);)
                       break;
                   }
               }    
            }            
            break;

       case choose:
            for( child = actionNode->firstChild;  child != NULL;
                 child = child->nextSibling) 
            {
                if (child->nodeType != ELEMENT_NODE) continue;
                switch (getTag(child)) {
                    case when:
                        str = getAttr(child, "test");
                        if (str) {
                            xs->current = currentNode;
                            rc = evalXPath(xs, context, currentNode, currentPos,
                                           str, &rs, errMsg);
                            CHECK_RC;
                            b = xpathFuncBoolean( &rs );
                            xpathRSFree( &rs ); 
                            if (b) {
                                /* process the children as well */
                                rc = ExecActions(xs, context, currentNode, currentPos,
                                                 child->firstChild, errMsg);
                                CHECK_RC;
                                return 0;
                            }                            
                        }
                        break;
                        
                    case otherwise:
                        /* process the children as well */
                        rc = ExecActions(xs, context, currentNode, currentPos,
                                         child->firstChild, errMsg);
                        CHECK_RC;
                        return 0;
                        
                    default:
                        *errMsg = strdup("only otherwise or when allowed in choose!");
                        return -1;
                }
            }
            break;
            
        case comment:
            pc = xpathGetTextValue (actionNode, &len);
            domAppendNewTextNode(xs->lastNode, pc, len, COMMENT_NODE);
            break;
            
        case copy:
            DBG(fprintf(stderr, "copy '%s' \n", currentNode->nodeName);)
            if (currentNode->nodeType == TEXT_NODE) {
                DBG(fprintf(stderr, "node is TEXT_NODE \n");)
                tnode = (domTextNode*)currentNode;
                savedLastNode = xs->lastNode;
                n = (domNode*)
                    domAppendNewTextNode(xs->lastNode, 
                                         tnode->nodeValue,
                                         tnode->valueLength,
                                         TEXT_NODE);
                xs->lastNode = n;
            } else
            if (currentNode->nodeType == ELEMENT_NODE) {
                DBG(fprintf(stderr, "node is ELEMENT_NODE \n");)
                savedLastNode = xs->lastNode;
                n = domAppendNewElementNode(xs->lastNode, 
                                            currentNode->nodeName,
                                            domNamespaceURI(currentNode) );
                xs->lastNode = n;
                str = getAttr(actionNode, "use-attribute-sets");
                if (str) {
                    rc = ExecUseAttributeSets (xs, context, currentNode, currentPos,
                                               str, errMsg);
                    CHECK_RC;
                }            
            } else 
            if (currentNode->nodeType == PROCESSING_INSTRUCTION_NODE) {
                savedLastNode = xs->lastNode;
                pi = (domProcessingInstructionNode*)currentNode;
                n = (domNode*)
                    domNewProcessingInstructionNode (savedLastNode->ownerDocument,
                                                     pi->targetValue,
                                                     pi->targetLength, 
                                                     pi->dataValue,
                                                     pi->dataLength);
                if (savedLastNode->parentNode) {
                    domAppendChild(savedLastNode->parentNode, n);
                } else {
                    domAppendChild(savedLastNode->ownerDocument->rootNode, n);
                }
                xs->lastNode = n;
                
            } else 
            if (currentNode->nodeType == ATTRIBUTE_NODE) {
                fprintf(stderr, "node is ATTRIBUTE_NODE, xsl:copy not implemented !!! \n");
            }

            /* process the children as well */
            rc = ExecActions(xs, context, currentNode, currentPos,
                             actionNode->firstChild, errMsg);
            CHECK_RC;
            break;
            
        case copyOf:            
            select = getAttr(actionNode, "select");
            if (select != NULL) {
                xs->current = currentNode;
                rc = evalXPath(xs, context, currentNode, currentPos, select,
                               &rs, errMsg);
                CHECK_RC;            
                TRACE1(" copyOf select='%s':\n", select);
                DBG(rsPrint(&rs));
                if (rs.type == NodeSetResult) {
                    for (i=0; i<rs.nr_nodes; i++) { 
                        if (rs.nodes[i]->nodeType == DOCUMENT_NODE) {
                            child = rs.nodes[i]->firstChild;
                            while (child) {
                                n = domCloneNode(child, 1);
                                domAppendChild(xs->lastNode, n);
                                child = child->nextSibling;
                            }
                        } else {
                            n = domCloneNode(rs.nodes[i], 1);
                            domAppendChild(xs->lastNode, n);
                        }
                    }
                } else
                if (rs.type == AttrNodeSetResult) {
                    for (i=0; i<rs.nr_nodes; i++) { 
                        attr = (domAttrNode*)rs.nodes[i];
                        domSetAttribute(xs->lastNode, 
                                        attr->nodeName, attr->nodeValue);
                    }
                } else {
                    str = xpathFuncString( &rs );
                    TRACE1("copyOf: xpathString='%s' \n", str);
                    domAppendNewTextNode(xs->lastNode, str, strlen(str), TEXT_NODE);
                    free(str);
                }
                xpathRSFree( &rs );               
            }
            break;
            
        case decimalFormat: return 0;
        
        case element:
            nsAT = getAttr(actionNode, "namespace");
            str  = getAttr(actionNode, "name");
            if (str) {
                rc = evalAttrTemplates( xs, context, currentNode, currentPos,
                                        str, &str2, errMsg);
                CHECK_RC;
                
                ns = NULL;
                if (nsAT) {
                     
                    rc = evalAttrTemplates( xs, context, currentNode, currentPos,
                                            nsAT, &ns, errMsg);
                    CHECK_RC;
                }
                
                xs->lastNode = domAppendNewElementNode (xs->lastNode, str2, ns);
                free(str2);
                str = getAttr(actionNode, "use-attribute-sets");
                if (str) {
                    TRACE1("use-attribute-sets = '%s' \n", str);
                    rc = ExecUseAttributeSets (xs, context, currentNode, currentPos,
                                               str, errMsg);
                    CHECK_RC;
                }
                /* process the children as well */
                rc = ExecActions(xs, context, currentNode, currentPos,
                                 actionNode->firstChild, errMsg);
                CHECK_RC;
            }
            break;
            
        case fallback: return 0;
            
        case forEach:
            select = getAttr(actionNode, "select");
            if (!select) select = "*";
            TRACE2("forEach select from '%s' %d:\n", currentNode->nodeName, currentNode->nodeNumber);
            DBG(
              if (currentNode->firstChild) {
                fprintf(stderr, "forEach select from child '%s' %d:\n", currentNode->firstChild->nodeName, currentNode->firstChild->nodeNumber);
              }
            )
            xpathRSInit( &nodeList );
            rsAddNode( &nodeList, currentNode );
            DBG(rsPrint( &nodeList ));
            xs->current = currentNode;
            rc = evalXPath(xs, &nodeList, NULL, 0, select, &rs, errMsg);
            CHECK_RC;
            TRACE1("forEach: evalXPath for select = '%s' gave back:\n", select);
            DBG(rsPrint(&rs));

            if ((rs.type == NodeSetResult) || (rs.type == AttrNodeSetResult)) {

                rc = doSortActions (xs, &rs, actionNode, &sorted, errMsg);
                CHECK_RC;
                if (!sorted) sortByDocOrder(&rs);

                TRACE1(" forEach for select = '%s': (SORTED)\n", select);
                DBG(rsPrint(&rs));

                for (i=0; i<rs.nr_nodes; i++) {
                    /* process the children as well */
                    rc = ExecActions(xs, &rs, rs.nodes[i], i,
                                     actionNode->firstChild, errMsg);
                    CHECK_RC;
                }
            } else
           
            xpathRSFree( &rs );
            break;
            
        case xsltIf:
            str = getAttr(actionNode, "test");
            if (str) {
                xs->current = currentNode;
                rc = evalXPath(xs, context, currentNode, currentPos, str, 
                               &rs, errMsg);
                CHECK_RC;
                b = xpathFuncBoolean( &rs );
                xpathRSFree( &rs ); 
                if (b) {
                    /* process the children as well */
                    rc = ExecActions(xs, context, currentNode, currentPos,
                                     actionNode->firstChild, errMsg);
                    CHECK_RC;
                }
            }
            break;

        case import:
        case include:
        case key:
            return 0;
                        
        case message:
            str  = getAttr(actionNode,"terminate"); 
            if (!str)  str = "no";
            /* jcl: TODO */
            break;
            
        case namespaceAlias: return 0;
        
        case number:
            value = getAttr(actionNode,"value");
            level = getAttr(actionNode,"level");  if (!level) level = "single";
            count = getAttr(actionNode,"count");  if (!count) count = ".";
            str   = getAttr(actionNode,"format"); if (!str)   str   = "1";
            rc = evalAttrTemplates( xs, context, currentNode, currentPos,
                                    str, &format, errMsg);
            CHECK_RC;            
            rc = xsltNumber(xs, value, level, count, format, 
                            context, currentNode, currentPos, errMsg);
            CHECK_RC;
            free(format);                            
            break;
        
        case output:
        case otherwise:
            return 0;
             
        case param:
            str = getAttr(actionNode, "name");
            if (str) {
                TRACE1("setting param '%s' ??\n", str);
                if (!xsltVarExists(xs, str)) {
                    TRACE1("setting param '%s': yes \n", str);
                    select = getAttr(actionNode, "select");
                    if (select) {
                        TRACE1("param select='%s'\n", select);
                        rc = xsltSetVar(xs, 0, str, context, currentNode, currentPos, 
                                        select, NULL, errMsg);
                    } else {
                        rc = xsltSetVar(xs, 0, str, context, currentNode, currentPos,
                                        NULL, actionNode->firstChild, errMsg);
                    }
                    CHECK_RC;
                }
            }        
            break;
            
        case preserveSpace: return 0;
        
        case procinstr:
            str = getAttr(actionNode, "name");
            if (str) {
                rc = evalAttrTemplates( xs, context, currentNode, currentPos,
                                        str, &str2, errMsg);
                pc = xpathGetTextValue (actionNode, &len);
                n = (domNode*)domNewProcessingInstructionNode(
                                 xs->resultDoc, str2, strlen(str), pc, len);
                domAppendChild(xs->lastNode, n);
                free(str2);
            }
            break;
            
        case sort:
        case stylesheet: 
        case stripSpace:
        case template:
            return 0;
        
        case text:
            /* jcl: todo */
            disableEsc = getAttr(actionNode, "disable-output-escaping");
            pc = xpathGetTextValue (actionNode, &len);
            DBG(fprintf(stderr, "text: pc='%s'%d \n", pc, len);)
            domAppendNewTextNode(xs->lastNode, pc, len, TEXT_NODE);
            break;
            
        case transform: return 0;
        
        case valueOf:
            /* jcl: todo */
            disableEsc = getAttr(actionNode, "disable-output-escaping");
            str = getAttr(actionNode, "select");
            if (str) {
                TRACE1("valueOf: str='%s' \n", str);
                xs->current = currentNode;
                rc = evalXPath(xs, context, currentNode, currentPos, str,
                               &rs, errMsg);
                CHECK_RC;
                DBG(rsPrint(&rs));
                str = xpathFuncString( &rs );
                TRACE1("valueOf: xpathString='%s' \n", str);
                domAppendNewTextNode(xs->lastNode, str, strlen(str), TEXT_NODE);
                xpathRSFree( &rs ); 
                free(str);
            }
            break;
            
        case variable:
            str = getAttr(actionNode, "name");
            if (str) {
                select = getAttr(actionNode, "select");
                if (select) {
                    /*TRACE1("variable select='%s'\n", select);*/
                    rc = xsltSetVar(xs, 0, str, context, currentNode, currentPos, 
                                    select, NULL, errMsg);
                } else {
                    rc = xsltSetVar(xs, 0, str, context, currentNode, currentPos, 
                                    NULL, actionNode->firstChild, errMsg);
                }
                CHECK_RC;
            }        
            break;
            
        case when:
        case withParam:
            return 0;
            
        default:
            savedLastNode = xs->lastNode;
            fprintf(stderr,
                    "append new tag '%s' uri='%s' \n", actionNode->nodeName,
                                                       domNamespaceURI(actionNode) );
                                                                
            xs->lastNode  = domAppendNewElementNode(xs->lastNode,
                                                    actionNode->nodeName,
                                                    domNamespaceURI(actionNode) );
            n = xs->lastNode;
            
            /* process the attributes */
            attr = actionNode->firstAttr;
            while (attr != NULL) {
                uri = domNamespaceURI((domNode*)attr);
                if (uri && strcmp(uri, XSLT_NAMESPACE)==0) {
                    domSplitQName((char*)attr->nodeName, prefix, &localName);
                    if (strcmp(localName,"use-attribute-sets")==0) {
                        str = attr->nodeValue;
                        rc = ExecUseAttributeSets (xs, context, currentNode, currentPos,
                                                   str, errMsg);
                        CHECK_RC;
                    }
                } else {
                    rc = evalAttrTemplates( xs, context, currentNode, currentPos,
                                            attr->nodeValue, &str, errMsg);
                    CHECK_RC;
                    domSetAttributeNS (n, attr->nodeName, str, uri);
                    free(str);
                }
                attr = attr->nextSibling;
            } 

            /* process the children as well */
            rc = ExecActions(xs, context, currentNode, currentPos,
                             actionNode->firstChild, errMsg);
            CHECK_RC;
            xs->lastNode = savedLastNode;
            return 0;
    }
    return 0;
}


/*----------------------------------------------------------------------------
|   ExecActions
|
\---------------------------------------------------------------------------*/
static int ExecActions (
    xsltState       * xs,
    xpathResultSet  * context,
    domNode         * currentNode,
    int               currentPos,
    domNode         * actionNode,
    char           ** errMsg
)
{
    domNode *savedLastNode, *savedCurrentNode;
    int rc;
    
    savedLastNode    = xs->lastNode;
    savedCurrentNode = xs->current;
    
    while (actionNode) {
        xs->current = currentNode;
        rc = ExecAction (xs, context, currentNode, currentPos, actionNode, errMsg);
        CHECK_RC;
        actionNode = actionNode->nextSibling;
    }
    xs->lastNode = savedLastNode;
    xs->current  = savedCurrentNode;
    return 0;
}



/*----------------------------------------------------------------------------
|   ApplyTemplate
|
\---------------------------------------------------------------------------*/
int ApplyTemplate (
    xsltState      * xs,
    xpathResultSet * context,
    domNode        * currentNode,
    int              currentPos,
    char           * mode,
    char          ** errMsg
)
{
    xsltTemplate   * tpl;
    xsltTemplate   * tplChoosen;
    domNode        * child;
    xpathResultSet   rs;
    int              rc;
    double           currentPrio;


    TRACE2("\n\nApplyTemplate mode='%s' currentPos=%d \n", mode, currentPos);
    DBG(printXML (currentNode, 0, 1);)
    
    xsltPushVarFrame (xs);

    /*--------------------------------------------------------------
    |   find template
    \-------------------------------------------------------------*/
    tplChoosen  = NULL;
    currentPrio = -100000.0;
    
    
    for( tpl = xs->templates; tpl != NULL; tpl = tpl->next) {
    
        TRACE3("find tpl match='%s' mode='%s' name='%s'\n",
               tpl->match, tpl->mode, tpl->name);
        
        /* exclude those, which don't match the current mode */
        if (   ( mode && !tpl->mode)
            || (!mode &&  tpl->mode)
            || ( mode &&  tpl->mode && (strcmp(mode,tpl->mode)!=0))
        ) {
            TRACE("doesn't match mode\n");
            continue; /* doesn't match mode */
        }
        
        if (tpl->match) {
        
            TRACE2("testing XLocPath '%s' for node %d \n", tpl->match, currentNode->nodeNumber);

            rc = xpathMatches ( tpl->ast, currentNode, &(xs->cbs), errMsg);
            TRACE1("xpathMatches = %d \n", rc);
            if (rc < 0) {
                TRACE1("xpathMatches had errors '%s' \n", *errMsg);
                xsltPopVarFrame(xs);
                return rc;
            }
            if (rc == 0) continue;
            TRACE1("matching '%s' \n", tpl->match);
            if (tpl->prio > currentPrio) {
                currentPrio = tpl->prio;
                tplChoosen = tpl;
                TRACE1("TAKING '%s' \n", tpl->match);
            }
        }
    }

    if (tplChoosen == NULL) {

        TRACE("nothing matches -> execute built-in template \n");

        /*--------------------------------------------------------------------
        |   execute built-in template
        \-------------------------------------------------------------------*/
        if (currentNode->nodeType == TEXT_NODE) {
            domAppendNewTextNode(xs->lastNode,
                                 ((domTextNode*)currentNode)->nodeValue,
                                 ((domTextNode*)currentNode)->valueLength, 
                                 TEXT_NODE);
            return 0;
        } else
        if (currentNode->nodeType == DOCUMENT_NODE) {
            child = ((domDocument*)currentNode)->documentElement;
        } else
        if (currentNode->nodeType == ELEMENT_NODE) {
            child = currentNode->firstChild;
        } else {
            return 0; /* for all other node we don't have to recurse deeper */
        }
        xpathRSInit( &rs );
        while (child) {    
            rsAddNode( &rs, child);
            child = child->nextSibling; 
        }
        rc = ApplyTemplates(xs, &rs, mode, errMsg);
        CHECK_RC;
        xpathRSFree( &rs );


    } else {

        TRACE1("tplChoosen '%s' \n", tplChoosen->match);
        DBG(printXML (tplChoosen->content->firstChild, 0, 1);)
        rc = ExecActions(xs, context, currentNode, currentPos, 
                         tplChoosen->content->firstChild, errMsg);
        TRACE1("ApplyTemplate/ExecActions rc = %d \n", rc);
        CHECK_RC;    
    }
    xsltPopVarFrame(xs);
    return 0;
}


/*----------------------------------------------------------------------------
|   ApplyTemplates
|
\---------------------------------------------------------------------------*/
int ApplyTemplates (
    xsltState      * xs,
    xpathResultSet * nodeList,
    char           * mode,
    char          ** errMsg
)
{
    domNode *savedLastNode;
    int      i, rc;
        
    if ((nodeList->type == NodeSetResult) || 
        (nodeList->type == AttrNodeSetResult)
    ) {
        savedLastNode = xs->lastNode;
        for (i=0; i < nodeList->nr_nodes; i++) {
            rc = ApplyTemplate (xs, nodeList, nodeList->nodes[i], i, mode, errMsg);
            CHECK_RC;
        }
        xs->lastNode = savedLastNode;            
    } else {
        TRACE("ApplyTemplates: nodeList not a NodeSetResult !!!\n");
        DBG(rsPrint(nodeList);)
    }
    return 0;
}


/*----------------------------------------------------------------------------
|   fillElementList
|
\---------------------------------------------------------------------------*/
void fillElementList (
    int    * nr_elems,
    char *** elems,
    char   * str
)
{
    char *pc, *start, save;
    int  count;
    
    count = 0;
    pc = str;
    while (*pc) {
        while (*pc == ' ') pc++;
        if (*pc == '\0') break;
        while (*pc && (*pc != ' ')) pc++;
        count++;
    }
    *nr_elems = count;
    *elems = (char**)malloc(sizeof(char*) * count);

    count = 0;
    pc = str;
    while (*pc) {
        while (*pc == ' ') pc++;
        start = pc;
        if (*pc == '\0') break;
        while (*pc && (*pc != ' ')) pc++;
        save = *pc;
        *pc = '\0';
        (*elems)[count] = strdup(start);
        count++;
        *pc = save;                   
    }
}    


/*----------------------------------------------------------------------------
|   StripSpace
|
\---------------------------------------------------------------------------*/
void StripSpace (
    xsltState  * xs,
    domNode    * node
)
{
    domNode *child, *newChild, *parent;
    int     i, len, onlySpace;
    char   *p;
    
    if (node->nodeType == TEXT_NODE) {
        p = ((domTextNode*)node)->nodeValue;
        len = ((domTextNode*)node)->valueLength;
        onlySpace = 1;
        for (i=0; i<len; i++) {           
            if ((*p!=' ') && (*p!='\n') && (*p!='\r') && (*p!='\t')) {
                onlySpace = 0;
                break;
            }
            p++;
        }
        if (onlySpace) {
            if (node->parentNode && (getTag(node->parentNode) == text)) {
                /* keep white texts below xsl:text elements */
                return;
            }
            parent = node->parentNode;
            while (parent) {                
                p = getAttr(parent,"space");
                if (p!=NULL) {
                    if (strcmp(p,"preserve")==0) return;
                    if (strcmp(p,"default")==0)  break;
                }
                parent = parent->parentNode;
            }
            DBG(fprintf(stderr, "removing %d(len %d) under '%s' \n", node->nodeNumber, len, node->parentNode->nodeName);)
            domDeleteNode(node);
        }
    }
    if (node->nodeType == ELEMENT_NODE) {
        child = node->firstChild;
        while (child) {
            newChild = child->nextSibling;
            StripSpace (xs, child);
            child = newChild;
        }
    }
}


/*----------------------------------------------------------------------------
|   xsltProcess
|
\---------------------------------------------------------------------------*/
int xsltProcess (
    domDocument       * xsltDoc,
    domNode           * xmlNode,
    xpathFuncCallback   funcCB,
    void              * clientData,
    char             ** errMsg,
    domDocument      ** resultDoc   
)
{ 
    xpathResultSet  nodeList;
    xsltAttrSet    *attrSet;
    domNode        *node;
    int             rc;
    xsltState       xs;
    char           *str, *select,  *name, *match, *use;
    xsltKeyInfo    *keyInfo;
    
    *errMsg = NULL;  
    
 /* jcl: HACK!!! */
    if (xmlNode->parentNode == NULL) {
        /*xmlNode = (domNode*)xmlNode->ownerDocument;*/
        xmlNode = xmlNode->ownerDocument->rootNode;
    }
    DBG(printXML(xmlNode, 0, 1);)

    xs.cbs.varCB           = xsltGetVar;
    xs.cbs.varClientData   = (void*)&xs;
    xs.cbs.funcCB          = xsltXPathFuncs;
    xs.cbs.funcClientData  = &xs;
    xs.orig_funcCB         = funcCB;
    xs.orig_funcClientData = clientData;
    xs.varFrames           = NULL;
    xs.templates           = NULL;
    xs.resultDoc           = domCreateDoc();
    xs.xmlRootNode         = xmlNode;
    xs.lastNode            = xs.resultDoc->rootNode;
    xs.attrSets            = NULL;
    xs.keyInfos            = NULL;
    xs.nr_perserveSpace    = 0;
    xs.perserveSpace       = NULL;
    xs.nr_stripSpace       = 0;
    xs.stripSpace          = NULL;
    xs.outputMethod        = NULL;
    xs.outputEncoding      = NULL;
    xs.outputMediaType     = NULL;
    Tcl_InitHashTable ( &(xs.xpaths), TCL_STRING_KEYS);
    
    xsltPushVarFrame(&xs);
    xpathRSInit( &nodeList );
    rsAddNode( &nodeList, xmlNode); 

    node = xsltDoc->documentElement;

    if ((getTag(node) != stylesheet) && (getTag(node) != transform)) {
        TRACE("no stylesheet node found --> ExecAction ");
        rc = ExecAction (&xs, &nodeList, xmlNode, 1, node, errMsg);
        DBG(fprintf(stderr, "ExecAction: rc=%d \n", rc);)
        CHECK_RC;
        xsltPopVarFrame (&xs);
        xpathRSFree( &nodeList );
        xs.resultDoc->documentElement = xs.resultDoc->rootNode->firstChild;
        *resultDoc = xs.resultDoc;
        return 0;
    } 

    node = node->firstChild;
    while (node) {
        switch ( getTag(node) ) {
        
            case attributeSet:
                str = getAttr(node, "name");
                if (str) {
                    attrSet = (xsltAttrSet*)malloc(sizeof(xsltAttrSet));
                    attrSet->next    = xs.attrSets;
                    attrSet->content = node;
                    attrSet->name    = str;
                    xs.attrSets = attrSet;
                }
                break;

            case param:
                str = getAttr(node, "name");
                if (str) {
                    TRACE1("setting param '%s' ??\n", str);
                    if (!xsltVarExists(&xs, str)) {
                        TRACE1("setting param '%s': yes \n", str);
                        select = getAttr(node, "select");
                        if (select) {
                            TRACE1("param select='%s'\n", select);
                            rc = xsltSetVar(&xs, 0, str, &nodeList, xmlNode, 0, 
                                            select, NULL, errMsg);
                        } else {
                            rc = xsltSetVar(&xs, 0, str, &nodeList, xmlNode, 0,
                                            NULL, node->firstChild, errMsg);
                        }
                        CHECK_RC;
                    }
                }        
                break;
                
            case decimalFormat:
                break;

            case import:
                break;
                
            case include:
                break;
                
            case key:
                name = getAttr(node, "name");
                if (!name) {
                    *errMsg = strdup("xsl:key requires name attribute!");
                    return -1;
                }
                match = getAttr(node, "match");
                if (!match) {
                    *errMsg = strdup("xsl:key requires match attribute!");
                    return -1;
                }
                use = getAttr(node, "use");
                if (!use) {
                    *errMsg = strdup("xsl:key requires use attribute!");
                    return -1;
                }
                keyInfo = malloc(sizeof(xsltKeyInfo));

                keyInfo->name      = name;
                keyInfo->match     = match;
                keyInfo->use       = use;
                keyInfo->evaluated = 0;
                keyInfo->next      = xs.keyInfos;
                keyInfo->values    = NULL;
                xs.keyInfos = keyInfo;
                break;
                
            case namespaceAlias:
                break;
                
            case output:
                str = getAttr(node, "method");
                if (str) { xs.outputMethod    = strdup(str); }
                str = getAttr(node, "encoding");
                if (str) { xs.outputEncoding  = strdup(str); }
                str = getAttr(node, "media-type");
                if (str) { xs.outputMediaType = strdup(str); }
                str = getAttr(node, "doctype-public");
                str = getAttr(node, "doctype-system");
                break;
                
            case preserveSpace:
                str = getAttr(node, "elements");
                if (str) {
                    fillElementList(&xs.nr_perserveSpace, &xs.perserveSpace, str);
                }
                break;
                
            case stripSpace:
                str = getAttr(node, "elements");
                if (str) {
                    fillElementList(&xs.nr_stripSpace, &xs.stripSpace, str);
                }
                break;
                
            case template:
                rc = xsltAddTemplate (&xs, node, errMsg);
                CHECK_RC;
                break;
                
            case variable:
                str = getAttr(node, "name");
                if (str) {
                    select = getAttr(node, "select");
                    if (select) {
                        rc = xsltSetVar(&xs, 0, str, &nodeList, xmlNode, 0, 
                                        select, NULL, errMsg);
                    } else {
                        rc = xsltSetVar(&xs, 0, str, &nodeList, xmlNode, 0,
                                        NULL, node->firstChild, errMsg);
                    }
                    CHECK_RC;
                }        
                break;
                
            default:
                break;
        }
        node = node->nextSibling;
    }
    StripSpace (&xs, xsltDoc->documentElement);
    StripSpace (&xs, xmlNode);
    
    
    rc = ApplyTemplates(&xs, &nodeList, NULL, errMsg);
    if (rc != 0) {
        /* TODO: free xpaths */
        return rc;
    }

    xsltPopVarFrame (&xs);
    xpathRSFree( &nodeList );

    xs.resultDoc->documentElement = xs.resultDoc->rootNode->firstChild;
    *resultDoc = xs.resultDoc;

    /* TODO: free xpaths */
    return 0;

} /* xsltProcess */

