#!./tcldomsh

proc printNode { node indent } {
    if {[$node nodeType] != "ELEMENT_NODE"} return
    puts stderr "$indent [$node nodeName] '[$node prefix]'([$node namespaceURI]):[$node localName] -[$node attributes]-"
    foreach child [$node childNodes] {
        printNode $child "    $indent" 
    }
}
proc printXML { doc } {

    set root [$doc documentElement]

    printNode $root " "
    puts stderr ""
}

proc testNS { xml } {

    puts stderr ""
    puts stderr "-----------------------------------------------"
    puts stderr $xml
    puts stderr "-----------------------------------------------"
    set doc [dom parse $xml]
    printXML $doc 
}

testNS {<x xmlns:edi="http://ecommerce.org/schema"><a/></x>}
testNS {<x xmlns="http://ecommerce.org/schema"><price units='Euro'>32.18</price></x>} 
testNS \
{<x xmlns:edi='http://ecommerce.org/schema'>
    <!-- the 'taxClass' attribute's namespace is http://ecommerce.org/schema -->
    <lineItem edi:taxClass="exempt">Baby food</lineItem>
 </x>}
testNS \
{<x xmlns:edi="http://ecommerce.org/schema">
    <edi:lineItem taxClass="exempt">Baby food</edi:lineItem>
</x>}
testNS {<html:html xmlns:html="http://www.w3.org/TR/REC-hmtl40"><html:head><html:title>Fro</html:title></html:head></html:html>}
testNS \
{<?xml version="1.0"?>
<!-- all elements here are explicitly in the HTML namespace -->
<html:html xmlns:html='http://www.w3.org/TR/REC-html40'>
   <html:head><html:title>Frobnostication</html:title></html:head>
   <html:body><html:p>Moved to 
   <html:a href='http://frob.com'>here.</html:a></html:p></html:body>
</html:html>}

testNS {<bk:book xmlns:bk="urn:loc.gov:books" xmlns:isbn="urn:ISBN:0-3"><bk:title>Cheader</bk:title><isbn:number>22222</isbn:number></bk:book>}
testNS \
{<?xml version="1.0"?>
<!-- both namespace prefixes are available throughout -->
<bk:book xmlns:bk='urn:loc.gov:books'
         xmlns:isbn='urn:ISBN:0-395-36341-6'>
    <bk:title>Cheaper by the Dozen</bk:title>
    <isbn:number>1568491379</isbn:number>
</bk:book>}
 
testNS {<html xmlns='http://www.w3.org/TR/REC-html40'><head><title>Fro</title></head></html>}

testNS \
{<?xml version="1.0"?>
<!-- initially, the default namespace is "books" -->
<book xmlns='urn:loc.gov:books'
      xmlns:isbn='urn:ISBN:0-395-36341-6'>
    <title>Cheaper by the Dozen</title>
    <isbn:number>1568491379</isbn:number>
    <notes>
        <!-- make HTML the default namespace for some commentary -->
        <p xmlns='urn:w3-org-ns:HTML'>
            This is a <i>funny</i> book!
        </p>
    </notes>
</book>}

testNS \
{<?xml version='1.0'?>
<Beers>
    <!-- the default namespace is now that of HTML -->
    <table xmlns='http://www.w3.org/TR/REC-html40'>
     <th><td>Name</td><td>Origin</td><td>Description</td></th>
     <tr> 
       <!-- no default namespace inside table cells -->
       <td><brandName xmlns="">Huntsman</brandName></td>
       <td><origin xmlns="">Bath, UK</origin></td>
       <td>
         <details xmlns=""><class>Bitter</class><hop>Fuggles</hop>
           <pro>Wonderful hop, light alcohol, good summer beer</pro>
           <con>Fragile; excessive variance pub to pub</con>
         </details>
       </td>
     </tr>
     <tr>
     </tr>
    </table>
</Beers>}