<?xml version='1.0'?>
<!--
	$Id: tmml-html.xsl,v 2.24 2000/08/07 19:24:57 joe Exp $
	Author: 	Joe English, <jenglish@flightab.com>
	Created: 	26 Jun 2000
	Description:	Transform TMML 0.3 to HTML 
-->
<xsl:stylesheet
    xsl:version	= "1.0"
    xmlns:xsl	= "http://www.w3.org/1999/XSL/Transform"
>

<!-- Notes:
    Need to use 'encoding=us-ascii' to override the default UTF-8,
    which seems to make Netscape pick screwy fonts.

    'indent=yes' is helpful for debugging the output HTML,
    but it messes up PRE elements.

    &#183;	: ISO Latin 1 "middot" character
    &#160;	: ISO Latin 1 "non-breaking space" character
-->

<xsl:output method="html"
    doctype-public="-//W3C//DTD HTML 4.0 Transitional//EN"
    indent	= "no"
    encoding	= "us-ascii"
/>

<xsl:variable name="navsep"> &#183; </xsl:variable>
<xsl:variable name="removeChars" select="' &#x9;&#xA;%*()[];'"/>

<xsl:variable name="INDEX" select="string('document INDEX.MAP')"/>
<xsl:key name="INDEX" match="DEF" use="@name"/>

<xsl:variable name="debug" select="0"/>

<xsl:strip-space
    elements	= "section namesection synopsis keywords"
/>
<xsl:preserve-space
    elements 	= "example syntax HEAD"
/>

<!--  HTML header and footer: 
-->

<xsl:template name="header">
    <DIV class="navbar" align="center">
	<xsl:for-each select="namesection"
	    ><xsl:call-template name="section-xref"
	/></xsl:for-each>
	<xsl:for-each select="section | synopsis | seealso | keywords"
	    ><xsl:value-of select="$navsep"
	    /><xsl:call-template name='section-xref'
    /></xsl:for-each>
    </DIV>
    <HR class="navsep"/>
</xsl:template>

<xsl:template name="footer">
    <HR class="navsep"/>
    <!-- TODO:  put copyright information, etc. here -->
    <xsl:copy-of select="string('document footer.xml')"/>
</xsl:template>

<!-- Top-level rule:
-->
<xsl:template match="/manpage">
<HTML>
<HEAD>
    <LINK rel="stylesheet" href="manpage.css"/>
    <TITLE><xsl:value-of select="@title"/></TITLE>
    <META name="xsl-processor" content="{system-property('xsl:vendor')}"/>
    <META name="generator" content="$Id: tmml-html.xsl,v 2.24 2000/08/07 19:24:57 joe Exp $"/>
</HEAD>
<BODY>
<DIV class="header"><xsl:call-template name="header"/></DIV>
<DIV class="body"  ><xsl:call-template name="body"  /></DIV>
<DIV class="footer"><xsl:call-template name="footer"/></DIV>
</BODY>
</HTML>
</xsl:template>

<xsl:template name="body">
    <xsl:apply-templates/>
</xsl:template>


<!-- The 'class-attribute' template copies the GI of the source element
     to the 'class' attribute of the result element; this is used
     to preserve at least some semantic information in the output HTML.
-->
<xsl:template name="class-attribute"
    ><xsl:attribute name="class"
	><xsl:choose
	    ><xsl:when test="@cat"	><xsl:value-of select="name()"
	    /></xsl:when
	    ><xsl:otherwise 		><xsl:value-of select="name()"
	    /></xsl:otherwise
	></xsl:choose
    ></xsl:attribute
></xsl:template>

<!-- Standard front-matter sections:
-->
<xsl:template match="namesection"
    ><h2><a name="SECT{generate-id()}">NAME</a></h2>
    <p class="namesection"
    ><B class="names"><xsl:for-each select="name"
	><xsl:apply-templates
	/><xsl:if test="following-sibling::name">, </xsl:if
    ></xsl:for-each
    > - </B><BR
    /><xsl:apply-templates select="desc/node()"/></p>
</xsl:template>

<xsl:template match="synopsis">
<h2><a name="SECT{generate-id()}">SYNOPSIS</a></h2>
<xsl:apply-templates/>
</xsl:template>

<!-- Standard back-matter sections:
-->
<xsl:template match="seealso">
    <h2><a name="SECT{generate-id()}">SEE ALSO</a></h2>
    <P class="seealso"
	><xsl:for-each select="*"
	    ><xsl:apply-templates select="."
	    /><xsl:if test="following-sibling::*">, </xsl:if
	></xsl:for-each
    ></P
></xsl:template>

<xsl:template match="keywords">
    <h2><a name="SECT{generate-id()}">KEYWORDS</a></h2>
    <P class="keywords"
	><xsl:for-each select="keyword"
	    ><xsl:apply-templates select="."
	    /><xsl:if test="following-sibling::keyword">, </xsl:if
	></xsl:for-each
    ></P
></xsl:template>

<xsl:template match="keyword">
    <A class	= "keyword"
    	href	= "keyword-index.html#KW-{translate(./text(),$removeChars,'')}"
    ><xsl:apply-templates
    /></A
></xsl:template>

<!-- Structural elements:
-->
<xsl:template match="section">
    <xsl:apply-templates/>
</xsl:template>
<xsl:template match="section/title">
<h2><a name="SECT{generate-id(..)}"><xsl:apply-templates/></a></h2>
</xsl:template>

<xsl:template match="subsection">
    <xsl:apply-templates/>
</xsl:template>
<xsl:template match="subsection/title">
<h3><a name="SECT{generate-id(..)}"><xsl:apply-templates/></a></h3>
</xsl:template>

<xsl:template match="syntax[@scope='global' and @cat and @name]"
><A NAME="{@cat}_{@name}">&#160;</A>	<!-- !#@$ing Netscape -->
<PRE class="syntax" id="{@cat}_{@name}"
    ><xsl:apply-templates
/></PRE>
</xsl:template>

<xsl:template match="syntax | example"
    ><xsl:element name="PRE"
	><xsl:call-template name="class-attribute"
	/><xsl:apply-templates/>
    </xsl:element
></xsl:template>

<xsl:template match="arglist" priority="5">
<DIV class="arglist">
<!-- Without a DIV wrapper, Netscape misformats the table -->
<TABLE width="100%" rules="none" cellpadding="5%">
    <TR class="heading">
        <TH width="20%">Type</TH>
	<TH width="70%">Name</TH>
	<TH width="10%">Mode</TH>
    </TR>
<xsl:for-each select="argdef">
    <TR class="syntax">
	<TD class="type" width="20%" align="left"
	    ><xsl:apply-templates select="argtype"/></TD>
	<TD class="name" width="70%" align="left"
	    ><A name="ARG-{translate(name/text(),$removeChars,'')}"
	    ><xsl:apply-templates select="name/node()"/></A
	></TD>
	<TD class="mode" width="10%" align="center"
	    ><xsl:apply-templates select="argmode/node()"/></TD>
    </TR>
    <TR class="desc">
	<TD class="padding" width="20%">&#160;</TD>
	<TD class="argdesc" width="80%" align="left" colspan="2"
	    ><xsl:apply-templates select="desc/node()"/></TD>
    </TR>
</xsl:for-each>
</TABLE>
</DIV>
</xsl:template>

<xsl:template match="argtype">
    <xsl:variable name="href"
    	><xsl:call-template name="lookup-href"
	    ><xsl:with-param name="refname" select="."
	/></xsl:call-template
    ></xsl:variable>
    <xsl:choose>
    	<xsl:when test="$href"
	    ><a href="{$href}"><xsl:apply-templates/></a
	></xsl:when>
	<xsl:otherwise><xsl:apply-templates/></xsl:otherwise>
    </xsl:choose>
</xsl:template>

<xsl:template match="commandlist" priority="5">
<xsl:element name="dl">
    <xsl:attribute name="class">commandlist</xsl:attribute>
    <xsl:apply-templates/>
</xsl:element>
</xsl:template>

<xsl:template match="commanddef"
><xsl:apply-templates
/></xsl:template>

<xsl:template match="commanddef/command">
<xsl:element name="dt">
<xsl:apply-templates/>
</xsl:element>
</xsl:template>

<xsl:template match="commanddef/desc">
<xsl:element name="dd">
<xsl:apply-templates/>
</xsl:element>
</xsl:template>

<xsl:template match="optionlist[@cat]" priority="6">
    <xsl:call-template name="optionlist">
    	<xsl:with-param name="cat" select="@cat"/>
    </xsl:call-template>
</xsl:template>


<xsl:template name="optionlist" match="optionlist" priority="5">
<xsl:param name="cat" select="'option'"/>
<DIV class="table"><!-- %%% -->
<TABLE width="100%" rules="none" cellpadding="5%">
    <TR class="heading">
        <TH width="20%">Name</TH>
	<TH width="40%">Database name</TH>
	<TH width="40%">Database class</TH>
    </TR>
<xsl:for-each select="optiondef">
    <xsl:variable name="name"
	><xsl:choose
	    ><xsl:when test="name/@name"><xsl:value-of select="name/@name"
	    /></xsl:when
	    ><xsl:otherwise><xsl:value-of select="./name"
	    /></xsl:otherwise
	></xsl:choose
    ></xsl:variable>
    <TR class="syntax">
	<TD width="20%" align="left"
	    ><A name="{$cat}_{translate($name,$removeChars,'')}"
	    ><xsl:apply-templates select="name/node()"/></A></TD>
	<TD width="40%" align="left"
	    ><xsl:apply-templates select="dbname/node()"/></TD>
	<TD width="40%" align="left"
	    ><xsl:apply-templates select="dbclass/node()"/></TD>
    </TR>
    <TR class="desc">
	<TD class="padding" width="20%">&#160;</TD>
	<TD class="argdesc" width="80%" align="left" colspan="2"
	    ><xsl:apply-templates select="desc/node()"/></TD>
    </TR>
</xsl:for-each>
</TABLE>
</DIV>
</xsl:template>


<!-- Inline elements:
-->

<xsl:template match="meta|m"
    ><xsl:element name="i"
	><xsl:call-template name="class-attribute"
	/><xsl:apply-templates
    /></xsl:element
></xsl:template>

<xsl:template match="emph"
    ><xsl:element name="em"><xsl:apply-templates/></xsl:element
></xsl:template>

<xsl:template match="lit|l|samp"
    ><xsl:element name="tt"
	><xsl:call-template name="class-attribute"
	/><xsl:apply-templates
    /></xsl:element
></xsl:template>

<xsl:template 
    match="cmd|variable|method|syscmd|fun|option|widget|package|typedef"
    ><xsl:element name="b"
	><xsl:call-template name="class-attribute"
	/><xsl:call-template name="make-xref"
    /></xsl:element
></xsl:template>

<!-- Extensions: -->
<xsl:template match="term | *[@tmml='term']"
    ><xsl:element name="b"
	><xsl:call-template name="class-attribute"
	/><xsl:call-template name="make-xref"
    /></xsl:element
></xsl:template>

<!-- These map one-one onto HTML elements:
-->
<xsl:template match="p | dl | dt | dd | ul | ol | li"
    ><xsl:element name="{name()}"><xsl:apply-templates/></xsl:element
></xsl:template>

<xsl:template match="dl[@scope='global' and @cat]/dt">
    <xsl:variable name="cat"  select="parent::*/@cat"/>
    <xsl:variable name="name" select="."/>
    <xsl:element name="dt">
    	<A name="{$cat}_{$name}"><xsl:apply-templates/></A>
    </xsl:element>
</xsl:template>

<xsl:template match="br"><xsl:element name="br"/></xsl:template>
<xsl:template match="b | i"
    ><xsl:element name="{name()}"
    ><xsl:if test="@class"
        ><xsl:attribute name="class"
	    ><xsl:value-of select="@class"/></xsl:attribute
	></xsl:if
    ><xsl:apply-templates
    /></xsl:element
></xsl:template>

<!-- "Simple" lists:
-->

<xsl:template match="sl[@cols]" priority="2">
    <DIV class="menu">
    <TABLE class="menu" width="100%">
	<xsl:call-template name="slrow">
	    <xsl:with-param name="cols" select="number(@cols)"/>
	    <xsl:with-param name="cat"  select="@cat"/>
	</xsl:call-template>
    </TABLE>
    </DIV>
</xsl:template>

<xsl:template match="sl" priority="1">
    <xsl:element name="ul">
	<xsl:apply-templates/>
    </xsl:element>
</xsl:template>

<xsl:template name="slrow">
    <xsl:param name="cols"/>
    <xsl:param name="start" select="1"/>
    <xsl:param name="end"   select="$start + $cols"/>
    <xsl:param name="cells" select="count(li)"/>
    <xsl:param name="colwidth" select="floor(100 div $cols)"/>
    <xsl:element name="TR">
        <xsl:for-each 
	    select="li[position() &gt;= $start and position() &lt; $end]"
	>
	    <TD width="{$colwidth}%"
		><xsl:apply-templates select="."
	    /></TD>
	</xsl:for-each>
    </xsl:element>
    <xsl:if test="$end &lt;= $cells">
	<xsl:call-template name="slrow">
	    <xsl:with-param name="cols"	 	select="$cols"/>
	    <xsl:with-param name="start"	select="$end"/>
	</xsl:call-template>
    </xsl:if>
</xsl:template>

<xsl:template match="sl[@cat]/li" priority="3">
    <xsl:call-template name="make-xref"/>
</xsl:template>

<xsl:template match="sl/li" priority="2">
    <xsl:apply-templates/>
</xsl:template>

<!-- XL ("extended" lists)
--> 
<xsl:template match="xl | *[@tmml='xl']">
    <DIV class="table">
    <TABLE width="100%" rules="none" cellspacing="0%">
	<xsl:if test="@xltitles">
	    <TR class="heading">
		<xsl:call-template name="xl-titles">
		    <xsl:with-param name="titles" 
		    	select="concat(@xltitles,'/')"/>
		</xsl:call-template>
	    </TR>
	</xsl:if>
	<xsl:for-each select="child::xle | child::*[@tmml='xle']">
	    <TR class="syntax">
		<xsl:apply-templates mode="xle-defs"/>
	    </TR>
	    <TR class="desc">
		<!-- <TD class="padding">&#160;</TD> -->
		<TD class="desc" 
			colspan="{count(child::desc/preceding-sibling::*)}">
		    <xsl:apply-templates select="desc/node()"/>
		</TD>
	    </TR>
    	</xsl:for-each>
    </TABLE>
    </DIV>
</xsl:template>

<xsl:template mode="xle-defs" match="desc" priority="2"/>
<xsl:template mode="xle-defs" match="name" priority="2">
    <TD><A name="{ancestor::*[@cat][1]/@cat}_{.}"
	    ><xsl:apply-templates/></A
    ></TD>
</xsl:template>
<xsl:template mode="xle-defs" match="*" priority="1">
    <TD><xsl:apply-templates/></TD>
</xsl:template>

<xsl:template name="xl-titles">
    <xsl:param name="titles"/>
    <xsl:variable name="first" select="substring-before($titles, '/')"/>
    <xsl:variable name="rest"  select="substring-after($titles, '/')"/>
    <xsl:if test="$first">
	<TH><xsl:value-of select="$first"/></TH>
	<xsl:call-template name="xl-titles">
	    <xsl:with-param name="titles" select="$rest"/>
	</xsl:call-template>
    </xsl:if>
</xsl:template>

<!-- Cross-references:
    template section-xref makes an internal cross-reference
    to the current node, which must be a section element.

    template make-xref makes a cross-reference from the current node,
    based on GI, category, etc.
-->

<xsl:template name="section-xref"
    ><xsl:variable name="sectname"
    ><xsl:call-template name="section-name"
    /></xsl:variable
    ><A HREF="#SECT{generate-id()}"
	><xsl:value-of select="$sectname"
    /></A
></xsl:template>

<xsl:template name="section-name"
><xsl:choose>
    <xsl:when test="self::namesection"	>NAME</xsl:when>
    <xsl:when test="self::synopsis"	>SYNOPSIS</xsl:when>
    <xsl:when test="self::seealso"	>SEE ALSO</xsl:when>
    <xsl:when test="self::keywords"	>KEYWORDS</xsl:when>
    <xsl:when test="@title"
    	><xsl:value-of select="@title"/></xsl:when>
    <xsl:when test="title"
	><xsl:apply-templates select="./title/node()"/></xsl:when>
    <xsl:otherwise>
	><xsl:message>Section title missing</xsl:message
	><xsl:value-of select="name()"
    /></xsl:otherwise>
</xsl:choose
></xsl:template>

<!-- BEGIN TEST -->
<xsl:template name="extract-refname"
><xsl:choose
    ><xsl:when test="contains(.,'(n)') or contains(.,'(3)') or contains(.,'(1)')"
	><xsl:value-of select="substring-before(.,'(')"
    /></xsl:when
    ><xsl:otherwise
    	><xsl:value-of select="."
    /></xsl:otherwise
></xsl:choose
></xsl:template>

<!-- %%% Do something about multiple matches. -->
<!-- %%% Check categories -->
<!-- %%% GAAAH! Rewrite this whole thing; it's a mess. -->

<xsl:template name="lookup-href"
    ><xsl:param name="refname"
    /><xsl:param name="from" select="ancestor::manpage/@id"
    /><xsl:for-each select="$INDEX"
    ><xsl:for-each select="key('INDEX',$refname)[1]"
	    ><xsl:value-of select="@manpage"/><xsl:text>.html</xsl:text
	><xsl:if test="@subpart"
	    ><xsl:text>#</xsl:text
	    ><xsl:value-of select="@subpart"
	/></xsl:if
    ></xsl:for-each
    ></xsl:for-each
></xsl:template>

<xsl:template match="ref[@refid]" priority="2">
    <xsl:variable name="tgtid" select="generate-id(key('id',@refid))"/>
    <A HREF="#SECT{$tgtid}"><xsl:apply-templates/></A>
</xsl:template>

<!-- %%% TO DO: check 'cat' attribute as well as refname. -->
<xsl:template name="make-xref" match="ref">
    <xsl:variable name="refname"
    	><xsl:call-template name="extract-refname"/></xsl:variable>
    <xsl:variable name="href"
    	><xsl:call-template name="lookup-href"
	    ><xsl:with-param name="refname" select="$refname"
	/></xsl:call-template
    ></xsl:variable>
    <xsl:choose>
	<!-- Don't make links to the same manpage -->
    	<xsl:when test="$href and $href != concat(ancestor::manpage/@id,'.html')">
	    <A href="{$href}"><xsl:apply-templates/></A>
	</xsl:when>
	<xsl:when test="$href">
	    <xsl:apply-templates/>
	</xsl:when>
	<xsl:otherwise>
	    <xsl:message>Missing link: <xsl:value-of select="."/> / <xsl:value-of select="$refname"
	    /> from <xsl:value-of select="ancestor::manpage/@id"/></xsl:message>
	    <xsl:apply-templates/>
	</xsl:otherwise>
    </xsl:choose>
</xsl:template>
<!-- END TEST -->


<!-- Cross-references:
 -->
<xsl:key name="id" match="*[@id]" use="@id"/>
<xsl:key name="arg" match="arglist/argdef/name[@name]" use="@name"/>
<xsl:key name="arg" match="arglist/argdef/name[not(@name)]"
	use="translate(./text(),' &#x9;&#xA;%*()[];','')"/> <!-- removeChars -->
<xsl:template priority="2"
	match="synopsis/syntax/m[key('arg',
		translate(./text(),' &#x9;&#xA;*()[]' ,''))]"
    ><A href="#ARG-{translate(./text(),$removeChars,'')}"
    	><i><xsl:apply-templates/></i></A>
</xsl:template>

<!-- Default catch-all rule, used to catch source elements
     for which there is no matching template 
-->
<xsl:template match="*"
    ><xsl:message terminate="no">Unrecognized GI <xsl:value-of select="name()"
    /></xsl:message
    ><xsl:choose
	><xsl:when test="$debug"
	    ><xsl:element name="EM"
		><xsl:attribute name="class">UNRECOGNIZED</xsl:attribute
		><xsl:processing-instruction name="UNRECOGNIZED"
		    ><xsl:value-of select="name()"
		/></xsl:processing-instruction
		><xsl:apply-templates
	    /></xsl:element
	></xsl:when
	><xsl:otherwise><xsl:apply-templates/></xsl:otherwise
    ></xsl:choose
></xsl:template>

</xsl:stylesheet>
