#!./tcldomsh


namespace eval ::dom::xpathFunc {
}

namespace eval ::dom::xpathFuncHelper {
}

proc ::dom::xpathFuncHelper::coerce2number { type value } {
    switch $type {
        empty      { return 0 }
        number -
        string     { return $value }
        attrvalues { return [lindex $value 0] }
        nodes      { return [[lindex $value 0] selectNodes number()] }
        attrnodes  { return [lindex $value 1] }
    }
}

proc ::dom::xpathFuncHelper::coerce2string { type value } {
    switch $type {
        empty      { return 0 }
        number -
        string     { return $value }
        attrvalues { return [lindex $value 0] }
        nodes      { return [[lindex $value 0] selectNodes string()] }
        attrnodes  { return [lindex $value 1] }
    }
}


proc ::dom::xpathFunc::format-number { ctxNode pos nodeListType nodeList args } {

    #puts stderr "$ctxNode $pos '$nodeListType' list:'$nodeList' args='$args'"

    set argLen [llength $args]
    if { ($argLen != 4) && ($argLen != 6) } {
        return -code error "wrong number of args: format-number(node,typeString,?decimalFormat?)"
    }
    foreach { arg1Typ arg1Value arg2Typ arg2Value } $args break
    set num       [::dom::xpathFuncHelper::coerce2number $arg1Typ $arg1Value ]
    set formatStr [::dom::xpathFuncHelper::coerce2string $arg2Typ $arg2Value ]
    #puts stderr "num:'$num' formatStr:'$formatStr'"
    set before 0
    set after  0
    set dotSeen 0
    foreach char [split $formatStr ""] {
        #puts stderr "'$char'"
        if {$char == "#"} {
            if {$dotSeen} { incr after } else { incr before }
        }
        if {$char == "."} { set dotSeen 1 }
    }
    incr before $after
    incr before $dotSeen
    set numStr [format "%${before}.${after}f" $num]
    #puts stderr "$before $after '$numStr'"
    regsub {\.} $numStr {} numStr 

    set out ""
    foreach char [split $formatStr ""] {
        if {$char == "#"} {
            append out [string index $numStr 0]
            set numStr [string range $numStr 1 end]
        } else {
            append out $char
        }
    }
    return [list string $out]
}

proc ::dom::xpathFunc::system-property { ctxNode pos nodeListType nodeList args } {

    puts stderr "system-property $ctxNode $pos '$nodeListType' list:'$nodeList' args='$args'"
    return [list string 1.7]
}

proc ::dom::xpathFunc::unparsed-entity-uri { ctxNode pos nodeListType nodeList args } {

    puts stderr "unparsed-entity-uri $ctxNode $pos '$nodeListType' list:'$nodeList' args='$args'"
    foreach { arg1Typ arg1Value } $args break
    set str [::dom::xpathFuncHelper::coerce2string $arg1Typ $arg1Value ]
    return [list string $str]
}

proc ::dom::xpathFunc::document { ctxNode pos nodeListType nodeList args } {

    puts stderr "document $ctxNode $pos '$nodeListType' list:'$nodeList' args='$args'"
    foreach { arg1Typ arg1Value } $args break
    set str [::dom::xpathFuncHelper::coerce2string $arg1Typ $arg1Value ]
    puts stderr "!!!!!! document function called with arg='$str' !!!!!!"
    return [list string $str]
}



foreach file $argv {
    regsub {\.xsl} $file {} base
    set xslFile ${base}.xsl
    set xmlFile ${base}.xml
    set outFile ${base}.outtdom
    puts stderr "$xslFile $xmlFile $outFile"
  
    set fd  [open $xslFile ]
    set xml [read $fd ]
    close $fd
    set xsltDoc [dom parse -keepEmpties $xml] 
    set root [$xsltDoc documentElement]
    #puts stderr "xsltDoc:\n[$root asXML]"

    #set xsltDoc [dom parse $xml] 

    set fd  [open $xmlFile ]
    set xml [read $fd ]
    close $fd
    set xmlDoc [dom parse -keepEmpties $xml] 
    set root [$xmlDoc documentElement]

    puts stderr [time { $root xslt $xsltDoc resultDoc } 1]

    set fd [open $outFile w]
    if {$resultDoc != ""} {
        set r [$resultDoc documentElement]
        set t $r
        set allText 1
        while {$t != ""} {
            if {[$t nodeType] != "TEXT_NODE"} { 
                set allText 0
                break
            }
            set t [$t nextSibling]
        }
        if {$allText} {
            while {$r != ""} {
                puts $fd [$r nodeValue]
                set r [$r nextSibling]
            }
        } else {
            while {$r != ""} {
                puts $fd [$r asXML]
                set r [$r nextSibling]
            }
        }
    }
    close $fd
}

